"""Selection components for evolutionary algorithms."""

import random

from evoalgos.sorting import LexicographicSorting
from evoalgos.sorting import HyperVolumeContributionSorting


class Selection(object):
    """Abstract base class for selection operators."""

    def select(self, population, number, already_chosen=None):
        """Return `number` individuals from the population.

        Note that `population` is not modified.

        Parameters
        ----------
        population : iterable
            The individuals to select from.
        number : int
            The number of individuals to select.
        already_chosen : iterable of Individual, optional
            Definitely surviving individuals not included in `population`,
            but which might influence these ones.

        """
        population_copy = list(population)
        self.reduce_to(population_copy, number, already_chosen)
        return population_copy


    def reduce_to(self, population, number, already_chosen=None):
        """Reduce population size to `number` and return the rejected.

        This method is abstract. Here, `population` should be modified
        in-place.

        Parameters
        ----------
        population : iterable
            The population to reduce in size.
        number : int
            The number of individuals to reduce to.
        already_chosen : iterable of Individual, optional
            Definitely surviving individuals not included in `population`,
            but which might influence these ones.

        Returns
        -------
        rejected : list
            The removed individuals.

        """
        raise NotImplementedError()



class UniformSelection(Selection):
    """Choose individuals random uniformly from the population."""

    def select(self, population, number, already_chosen=None):
        """Return `number` randomly drawn individuals from population.

        Note that `population` is not modified.

        Parameters
        ----------
        population : iterable
            The individuals to select from.
        number : int
            The number of individuals to select.
        already_chosen : iterable of Individual, optional
            This parameter has no influence for this selection.

        """
        if number >= len(population):
            return population[:]
        else:
            return random.sample(population, number)


    def reduce_to(self, population, number, already_chosen=None):
        """Discard randomly selected members of the population and return them.

        Parameters
        ----------
        population : iterable
            The population to reduce in size.
        number : int
            The number of individuals to reduce to.
        already_chosen : iterable of Individual, optional
            This parameter has no influence for this selection.

        Returns
        -------
        rejected : list
            The removed individuals.

        """
        random.shuffle(population)
        rejected = population[number:]
        del population[number:]
        return rejected



class BackwardElimination(Selection):
    """Wrapper for other selection components."""

    def __init__(self, selection):
        """Constructor.

        Parameters
        ----------
        selection : Selection
            The selection component that is forced to backward elimination
            by this wrapper.

        """
        self.selection = selection


    def reduce_to(self, population, number, already_chosen=None):
        """Reduce population size to `number` and return the rejected.

        This method iteratively calls the respective method of the selection
        component, removing one individual per call.

        Parameters
        ----------
        population : iterable
            The population to reduce in size.
        number : int
            The number of individuals to reduce to.
        already_chosen : iterable of Individual, optional
            Definitely surviving individuals not included in `population`,
            but which might influence these ones.

        Returns
        -------
        rejected : list
            The removed individuals.

        """
        selection = self.selection
        num_rejected = len(population) - number
        rejected = []
        for _ in range(num_rejected):
            rejected_in_step = selection.reduce_to(population,
                                                   len(population) - 1,
                                                   already_chosen)
            rejected.extend(rejected_in_step)
        return rejected



class TruncationSelection(Selection):
    """Choose strictly the best individuals according to sorting component.

    The features depend heavily on the used sorting component. In
    particular, the selection is only deterministic and elitistic if the
    sorting component is, too.

    """
    def __init__(self, sorting_component=None):
        """Constructor.

        Parameters
        ----------
        sorting_component : SortingComponent, optional
            The sorting component that is used to establish a ranking of
            the individuals. Default is
            :class:`LexicographicSorting<evoalgos.sorting.LexicographicSorting>`.

        """
        Selection.__init__(self)
        self.sorting_component = sorting_component
        if sorting_component is None:
            self.sorting_component = LexicographicSorting()


    def reduce_to(self, population, number, already_chosen=None):
        """Reduce population size to `number` and return the rejected.

        Parameters
        ----------
        population : iterable
            The population to reduce in size.
        number : int
            The number of individuals to reduce to.
        already_chosen : iterable of Individual, optional
            Definitely surviving individuals not included in `population`,
            but which might influence these ones.

        Returns
        -------
        rejected : list
            The removed individuals.

        """
        sort = self.sorting_component.sort
        random.shuffle(population)
        # sort the population
        sort(population)
        rejected = population[number:]
        del population[number:]
        # return the worst individuals
        return rejected



class HyperVolumeContributionSelection(Selection):
    """Use the individuals' hypervolume contribution as criterion.

    This selection is designed for multi-objective problems. The selection
    criterion is the individuals' hypervolume contribution to the dominated
    hypervolume of the non-dominated front it belongs to.

    """
    def __init__(self, offsets=None,
                 prefer_boundary_points=True,
                 do_backward_elimination=True):
        """Constructor.

        Parameters
        ----------
        offsets : iterable, optional
            For calculating the hypervolume, a reference point is required.
            The reference point is typically calculated as the worst
            objective values in the set of individuals plus an offset
            vector, which can be specified here. Default offset is
            [1.0, ..., 1.0].
        prefer_boundary_points : bool, optional
            This flag only pertains to the two-objective case. If it is
            set to True, the two boundary points (but not their potentially
            existing duplicates) of a front are guaranteed to be retained.
        do_backward_elimination : bool, optional
            This argument only has influence if more than one individual
            is to be removed. Backward elimination means that in a greedy
            fashion, the worst individuals are removed one by one. In this
            implementation, the alternative is a 'super-greedy' approach,
            which removes the necessary number of individuals without
            recalculating the fitness of the other ones in between. Default
            is True (the former approach), which is also recommended,
            because it is more accurate. This option exists, because the
            results of the variant with :class:`BackwardElimination` are
            not 100% correct (due to reference point construction being
            triggered in every iteration) and because some time can be
            saved.

        """
        self.do_backward_elimination = do_backward_elimination
        self.offsets = offsets
        self.hv_sorting = HyperVolumeContributionSorting(None,
                                                         None,
                                                         prefer_boundary_points)


    def reduce_to(self, population, number, already_chosen=None):
        """Reduce population size to `number` and return the rejected.

        Parameters
        ----------
        population : iterable
            The population to reduce in size.
        number : int
            The number of individuals to reduce to.
        already_chosen : iterable of Individual, optional
            This parameter has no influence for this selection.

        Returns
        -------
        rejected : list
            The removed individuals.

        """
        # assume that individuals are already evaluated
        population_copy = list(population)
        if number >= len(population):
            return []
        else:
            # shortcuts
            hv_sorting = self.hv_sorting
            compute_fronts = hv_sorting.identify_groups
            # the reference point is obtained from the population
            reference_point = self.construct_ref_point(population, self.offsets)
            hv_sorting.reference_point = reference_point
            # check for easier special case
            is_2d = True
            for individual in population:
                is_2d &= len(individual.objective_values) == 2
            fronts = compute_fronts(population_copy)
            if self.do_backward_elimination:
                # if more than one individual is discarded, the contribution
                # is calculated again for each one, because neighbors can be
                # affected
                rejected = []
                while len(population_copy) > number:
                    last_front = fronts[-1]
                    # sort the last front by hypervolume contribution
                    if len(last_front) > 1:
                        random.shuffle(last_front)
                        if is_2d:
                            hv_sorting.sort_front_2d(last_front)
                        else:
                            hv_sorting.sort_front(last_front)
                    # concatenation of fronts is the now ordered population
                    del population_copy[:]
                    for front in fronts:
                        population_copy.extend(front)
                    # remove worst and update fronts to avoid recalculation
                    removed = population_copy.pop()
                    last_front.remove(removed)
                    rejected.append(removed)
                    if len(last_front) == 0:
                        fronts.pop()
                population[:] = population_copy
                return rejected
            else:
                hv_sorting.sort(population_copy)
                population[:] = population_copy[0:number]
                return population_copy[number:]


    def construct_ref_point(self, individuals, offsets=None):
        """Construct a reference point from the given individuals.

        Parameters
        ----------
        individuals : iterable
            The individuals whose objective values are considered. For
            each objective, the worst value is taken.
        offsets : iterable, optional
            Non-negative offsets to be added to the worst values. The
            default is [1.0, ..., 1.0].

        """
        dimensions = len(individuals[0].objective_values)
        if offsets is None:
            if self.offsets is None:
                offsets = [1.0] * dimensions
            else:
                offsets = self.offsets
        worst_values = list(individuals[0].objective_values)
        for individual in individuals:
            for i in range(dimensions):
                coordinate = individual.objective_values[i]
                if coordinate > worst_values[i]:
                    worst_values[i] = coordinate
        ref_point = [worst_values[i] + offsets[i] for i in range(dimensions)]
        return ref_point
